#include <stdio.h>
#include <string.h>
#include <stdint.h>

#include <libdragon.h>
#include <display.h>
#include <system.h>

#include "unf/usb.h"
#include "unf/debug.h"

#include "../common.h"
#include "../memmap.h"

#include "../gba_memory.h"
#include "../gba_cc_lut.h"

// Usually 59.72750057 Hz, unless GBC_RATE is overclocked (for 60FPS)
#define GBA_FPS ((float) GBC_BASE_RATE) / (308 * 228 * 4)

//static s16 audio_sample_buffer[(audio_samples_per_frame + 1) * 2];
static s16 audio_sample_buffer[2194];
static float audio_samples_per_frame   = (float)(GBA_SOUND_FREQUENCY) / (float)(GBA_FPS);
static float audio_samples_accumulator = 0.0f;

static int16_t g_joypads[JOYPAD_PORT_COUNT][RETRO_DEVICE_ID_JOYPAD_R+1];

u32 skip_next_frame = 0;
boot_mode selected_boot_mode = boot_game;

int dynarec_enable;
int sprite_limit = 1;

u32 idle_loop_target_pc = 0xFFFFFFFF;
u32 translation_gate_target_pc[MAX_TRANSLATION_GATES];
u32 translation_gate_targets = 0;

surface_t* display_surface = NULL;
size_t gba_screen_pixel_offset = 0;

static void audio_run(void)
{
   u32 samples_to_read;
   u32 samples_produced;

   /* audio_samples_per_frame is decimal;
    * get integer component */
   samples_to_read = (u32)audio_samples_per_frame;

   /* Account for fractional component */
   audio_samples_accumulator += audio_samples_per_frame -
         (float)samples_to_read;

   if (audio_samples_accumulator >= 1.0f)
   {
      samples_to_read++;
      audio_samples_accumulator -= 1.0f;
   }

   samples_produced = sound_read_samples(audio_sample_buffer, samples_to_read);
   audio_push(audio_sample_buffer, samples_produced, false);
}

static void allocate_video_surface() {
   display_surface = display_get();

   gba_screen_pixels = display_surface->buffer;
   gba_screen_pixels += gba_screen_pixel_offset;
}

static void video_run(void)
{
   if (display_surface != NULL) {
      display_show(display_surface);
   }
   allocate_video_surface();
}

static void setup_screen_geometry() {
   resolution_t res;
   filter_options_t filt;

#ifdef FULLSCREEN
   res.width = GBA_SCREEN_WIDTH;
   res.height = GBA_SCREEN_WIDTH * 3 / 4;
   filt = FILTERS_RESAMPLE_ANTIALIAS;
#else // integer scaled, centered
   res.width = 320;
   res.height = 240;
   filt = FILTERS_RESAMPLE;
#endif
   // TODO: do any of the interlace options give us anything like interframe blending?
   res.interlaced = INTERLACE_OFF;

   gba_screen_pitch = res.width;
   // center in screen
   gba_screen_pixel_offset = ((res.width - GBA_SCREEN_WIDTH) + (res.width * (res.height - GBA_SCREEN_HEIGHT))) / 2;

   display_init(res, DEPTH_16_BPP, 3, GAMMA_NONE, filt);
}

int16_t input_state(unsigned port, unsigned device, unsigned index, unsigned id) {
   if (port < JOYPAD_PORT_COUNT && device == RETRO_DEVICE_JOYPAD) {
      return g_joypads[port][id];
   } else {
      return 0;
   }
}

static void input_poll() {
   joypad_poll();
   JOYPAD_PORT_FOREACH(i) {
      joypad_buttons_t inp = joypad_get_buttons(i);
      g_joypads[i][RETRO_DEVICE_ID_JOYPAD_A] = inp.a;
      g_joypads[i][RETRO_DEVICE_ID_JOYPAD_B] = inp.b;
      g_joypads[i][RETRO_DEVICE_ID_JOYPAD_UP] = inp.d_up;
      g_joypads[i][RETRO_DEVICE_ID_JOYPAD_DOWN] = inp.d_down;
      g_joypads[i][RETRO_DEVICE_ID_JOYPAD_LEFT] = inp.d_left;
      g_joypads[i][RETRO_DEVICE_ID_JOYPAD_RIGHT] = inp.d_right;
      g_joypads[i][RETRO_DEVICE_ID_JOYPAD_START] = inp.start;
      g_joypads[i][RETRO_DEVICE_ID_JOYPAD_SELECT] = inp.z;
      g_joypads[i][RETRO_DEVICE_ID_JOYPAD_L] = inp.l;
      g_joypads[i][RETRO_DEVICE_ID_JOYPAD_R] = inp.r;
   }
}

void set_fastforward_override(bool fastforward) {}

int stdio_stderr_redirect(char* buf, unsigned len) {
   usb_write(DATATYPE_TEXT, buf, len);
   return write(2, buf, len);
}

int main() {
   // UNFLoader interface plumbing for flash carts
   stdio_t console_calls = { 0, stdio_stderr_redirect, 0 };
   //debug_initialize();
   usb_initialize();
   hook_stdio_calls(&console_calls);
   printf("UNFLoader debug initialized!\n");

   // ISViewer (i.e. ares-emu)
   debug_init_isviewer();
   printf("ISViewer debug initialized!\n");

   dfs_init(DFS_DEFAULT_LOCATION);

   setup_screen_geometry();
   allocate_video_surface();

   audio_init(GBA_SOUND_FREQUENCY, 4);
   joypad_init();

   // gpsp input.c implements this directly.
   retro_set_input_state(input_state);
   
   memcpy(bios_rom, open_gba_bios_rom, sizeof(bios_rom));

   init_gamepak_buffer();
   init_sound();

   memset(gamepak_backup, 0xff, sizeof(gamepak_backup));
   if (load_gamepak(NULL, "rom:/rom.gba", 0, 0) != 0)
   {
      printf("Could not load the game file.");
      return -1;
   }

   reset_gba();
   while(true) {
      input_poll();
      update_input();

      /* This runs just a frame */
      #ifdef HAVE_DYNAREC
      if (dynarec_enable)
         execute_arm_translate(execute_cycles);
      else
      #endif
      {
         /* Sticky bits only used in interpreter */
         clear_gamepak_stickybits();
         execute_arm(execute_cycles);
      }
   
      audio_run();
      video_run();
   }
}
